/**
 *  Copyright Thomas Allweyer 2014
 *  
 *  This source code accompagnies the book BPMS - Einfhrung in Business Process Management-Systeme
 *  by Thomas Allweyer
 *  
 *  This software is being provided "as is", without any warranty. The use of the software
 *  is at your own risk. The software has an experimental character and should only be used
 *  for experimental purposes. It cannot be guaranteed that this software is free from defects.
 *  
 *  You are free to use and to modify this software, provided you include a reference to the
 *  original author in any derivative work. 
 */
package bonitareporting;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import java.util.Properties;


import org.bonitasoft.engine.api.LoginAPI;
import org.bonitasoft.engine.api.ProcessAPI;
import org.bonitasoft.engine.api.TenantAPIAccessor;
import org.bonitasoft.engine.bpm.data.ArchivedDataInstance;
import org.bonitasoft.engine.bpm.process.ArchivedProcessInstance;
import org.bonitasoft.engine.bpm.process.ProcessDeploymentInfo;
import org.bonitasoft.engine.bpm.process.ProcessDeploymentInfoCriterion;
import org.bonitasoft.engine.bpm.process.ProcessInstance;
import org.bonitasoft.engine.bpm.process.ProcessInstanceCriterion;
import org.bonitasoft.engine.exception.BonitaException;
import org.bonitasoft.engine.session.APISession;

public class ProcessReport {
	private ProcessAPI processAPI;
	private String filenameOverview;
	private String filenameDetail;
	
	
	/* 
	 * Creates two csv-Files with reports of Bonita processes.
	 * 
	 * It has been tested with Bonita BPM Community version 6.2.5 under Windows 7.
	 * 
	 * It works with a local Bonita installation. The Bonita engine needs to be running 
	 * before this program is started.
	 * 
	 * A file "processreport.properties" must be provided in a class path directory, 
	 * containing the path of the local Bonita installation and the path of the output
	 * directory, e.g.:  
	 * 
	 *     bonitapath=C:/BonitaBPMCommunity-6.3.2/workspace/bonita
	 *     outputpath=C:/temp
	 * 
	 * 
	 */
	public static void main(String[] args) throws BonitaException {
		
	    try {
			Properties properties = new Properties();
			InputStream stream = ProcessReport.class.getClassLoader()
			        .getResourceAsStream("processreport.properties");
			properties.load(stream);
			stream.close();
			
			String bonita_path=properties.getProperty("bonitapath");		
	        System.setProperty("bonita.home", bonita_path);
	        
	        System.out.println(bonita_path);
	        
	        final LoginAPI loginAPI = TenantAPIAccessor.getLoginAPI();
	        APISession session = loginAPI.login("admin", "bpm");
	        System.out.println("Connected to process engine.");
	        
	        ProcessAPI processAPI = TenantAPIAccessor.getProcessAPI(session);  
	        
	        String path = properties.getProperty("outputpath");
	        String filenameProcessesReport = path + "\\ProzesseReport";
	        String filenameInstancesReport = path + "\\ProzessinstanzenReport";
	        
	        File processReportFile = new File(filenameProcessesReport + ".csv");
	        File instancesReportFile = new File(filenameInstancesReport + ".csv");
	        int counter = 0;
	        String counterString = "";
	        while(processReportFile.exists() || instancesReportFile.exists()){
	        	counter++;
	        	processReportFile = new File(filenameProcessesReport + counter + ".csv");
	        	instancesReportFile = new File(filenameInstancesReport + counter + ".csv");
	        	counterString = Integer.toString(counter);
	        }
	        
	        
	        new ProcessReport(processAPI, filenameProcessesReport + counterString + ".csv", filenameInstancesReport + counterString + ".csv").writeReport();       
	    } catch(IOException e)
	    	{
	    	     e.printStackTrace();
	    	}
	}

	
	
	public ProcessReport(ProcessAPI processAPI, String filenameOverview, String filenameDetail){
		this.processAPI = processAPI;
		this.filenameOverview = filenameOverview;
		this.filenameDetail = filenameDetail;
	}
	
	public void writeReport(){
		
		List<ProcessDeploymentInfo> processDeploymentInfos = processAPI.getProcessDeploymentInfos(0, 1000, ProcessDeploymentInfoCriterion.NAME_ASC);
		List<ProcessInstance> processInstances = processAPI.getProcessInstances(0, 10000, ProcessInstanceCriterion.NAME_ASC);
		List<ArchivedProcessInstance> archivedProcessInstances = processAPI.getArchivedProcessInstances(0, 10000, ProcessInstanceCriterion.NAME_ASC);
	
		SimpleDateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
		
        try
    	{
        	PrintWriter writerOverview = new PrintWriter (new FileWriter(filenameOverview));
        	PrintWriter writerDetail = new PrintWriter (new FileWriter(filenameDetail));
        	System.out.println("Writing Reports");
        
        	List<String> dataNames = new ArrayList<String>();
        	String dataNamesString = retrieveDataNames(archivedProcessInstances, dataNames);
        	
        	writeHeadlines(writerOverview, writerDetail, dataNamesString);
        
        	
    	    for(ProcessDeploymentInfo processinfo:processDeploymentInfos){
            	long currentProcessId = processinfo.getProcessId();
            	
            	int numberOfRunningInstances = countRunningInstances(processInstances, currentProcessId); 
            	
            	long shortestDuration = 0;
            	long longestDuration = 0;
            	long mediumDuration = 0;
            	long durationSum = 0;
            	int numberOfArchivedProcessInstancesOfThisProcess = 0;
            	
            	for(ArchivedProcessInstance archivedInstance:archivedProcessInstances){
            		if(archivedInstance.getProcessDefinitionId()==currentProcessId){
            			
            			Date startdate = archivedInstance.getStartDate();
            			Date enddate = archivedInstance.getEndDate();
            			
            			long starttimeInMilliseconds = startdate.getTime();
            			long endtimeInMilliseconds = enddate.getTime();
            			long durationInSeconds = (endtimeInMilliseconds-starttimeInMilliseconds)/1000;
            			
            			if(durationInSeconds<shortestDuration || numberOfArchivedProcessInstancesOfThisProcess==0){
            				shortestDuration = durationInSeconds;
            			}
            			if(durationInSeconds>longestDuration){
            				longestDuration = durationInSeconds;
            			}
            			durationSum = durationSum + durationInSeconds;
            			
            			String dataValuesString = retrieveDataValues(archivedInstance, dataNames);
            			
            			String startdateString = dateFormat.format(archivedInstance.getStartDate());
            			String enddateString = dateFormat.format(archivedInstance.getEndDate());
            			String durationString = formatduration(durationInSeconds);
            			String durationInDaysString = durationInDays(durationInSeconds);
    	    	
		    	    	writerDetail.println(
		    	    			removeUnwantedCharacters(archivedInstance.getName()) + ";" 
		    	    			+ archivedInstance.getRootProcessInstanceId() + "; "
		    	    			+ startdateString + ";"
		    	    			+ enddateString + ";"
		    	    			+ durationString + ";"
		    	    			+ durationInDaysString + ";"
		    	    			+ dataValuesString
		    	    	);
		    	    	
		    	    	numberOfArchivedProcessInstancesOfThisProcess++;
            		}
            	}
            	
            	
            	long totalNumberOfProcessInstancesOfThisProcess = numberOfRunningInstances + numberOfArchivedProcessInstancesOfThisProcess;
            	if(numberOfArchivedProcessInstancesOfThisProcess>0){
	            	mediumDuration = durationSum / numberOfArchivedProcessInstancesOfThisProcess;
            	}
            	            	            	
            	String shortestDurationString = formatduration(shortestDuration);
            	String mediumDurationString = formatduration(mediumDuration);
            	String longestDurationString = formatduration(longestDuration);
            	String shortestDurationInDays = durationInDays(shortestDuration);
            	String mediumDurationInDays = durationInDays(mediumDuration);
            	String longestDurationInDays = durationInDays(longestDuration);
            	
            	writerOverview.println(removeUnwantedCharacters(processinfo.getName()) + ";"
            			+ numberOfArchivedProcessInstancesOfThisProcess + ";"
            			+ numberOfRunningInstances + "; "
            			+ totalNumberOfProcessInstancesOfThisProcess + ";"
            			+ removeUnwantedCharacters(shortestDurationString) + ";"
            			+ removeUnwantedCharacters(mediumDurationString) + ";"
            			+ removeUnwantedCharacters(longestDurationString) +";"
            	        + removeUnwantedCharacters(shortestDurationInDays) + ";"
    			        + removeUnwantedCharacters(mediumDurationInDays) + ";"
    			        + removeUnwantedCharacters(longestDurationInDays)
    			  );
            	
    	    }
    	    
    	    writerDetail.close();
    	    writerOverview.close();
    	}
        catch(IOException e)
    	{
    	     e.printStackTrace();
    	}
        
        System.out.println("Finished writing Reports");
	}
	
		
	/*
	 * Returns a semicolon separated string with the archived process instance's
	 * data values. The position of each value is the position of the respective
	 * data object's name in the list dataNames. 
	 */
	private String retrieveDataValues(
			ArchivedProcessInstance archivedInstance,
			List<String> dataNames) {
		
		String valuesString = "";
		try{
			List<ArchivedDataInstance> datainstances = processAPI.getArchivedProcessDataInstances(
					                           archivedInstance.getRootProcessInstanceId(), 0, 1000);
			
			
			String[] values = new String[dataNames.size()];
			for(ArchivedDataInstance datainstance:datainstances){
				String datainstancename = datainstance.getName();
	
				int index = 0;
				Iterator<String> varIt = dataNames.iterator();
				while(varIt.hasNext()){
					String currentvarname = varIt.next();
					if(currentvarname.equals(datainstancename)){
						if(datainstance.getValue()!=null){
							values[index]=datainstance.getValue().toString();
						}
					}
					index++;
				}
			}
			for(String currentvalue:values){
				if(currentvalue!=null){
					valuesString = valuesString + removeUnwantedCharacters(currentvalue);
				}
				valuesString = valuesString + ";";
			}
		} catch (Exception e){
			System.out.println("Couldn't retrieve data values - user defined class used as type");
		}
		return valuesString;
	}

	
	/*
	 * Retrieves the names of all data (with simple data types) of all archived process instances and
	 * adds them to the list dataNames.
	 * Returns a semicolon-separated string of the data names.
	 */
	private String retrieveDataNames(List<ArchivedProcessInstance> archivedProcessInstances, List<String> dataNames) {
		String dataNamesString="";
		
		for(ArchivedProcessInstance archivedInstance:archivedProcessInstances){
			try{
				List<ArchivedDataInstance> datainstances = processAPI.getArchivedProcessDataInstances(archivedInstance.getRootProcessInstanceId(), 0, 1000);
				for(ArchivedDataInstance datainstance:datainstances){
					String varname = datainstance.getName();
					String vartype = datainstance.getClassName();
					if(vartype.contains("Integer")  || vartype.contains("String") || vartype.contains("Long") || 
							vartype.contains("Boolean") || vartype.contains("Double") || vartype.contains("Float")){
						boolean included = false;
						for(String dataName:dataNames){
							if(dataName.equals(varname)){
								included = true;
							}
						}
						if(!included){
							dataNames.add(varname);
						}
					}
				}
			} catch (Exception e){
				System.out.println("Couldn't retrieve data names - user defined class used as type");
			}
		}
				
		for (String entry:dataNames){
			dataNamesString = dataNamesString + "; " + entry;
		}
		
		return dataNamesString;
	}
	
	
    /* 
     * Returns the number of the process's running instances.
     */
	private int countRunningInstances(List<ProcessInstance> processInstances,
			long currentProcessId) {
    	int runningInstances = 0;
    	
    	for(ProcessInstance processinstance:processInstances){
    		if(processinstance.getProcessDefinitionId()==currentProcessId){
    			runningInstances++;
    		}
    	}
		return runningInstances;
	}
	
	
	/*
	 * Writes the headlines to both csv files.
	 * dataNamesString must contain all data names (semicolon-separated).
	 */
	private void writeHeadlines(PrintWriter writerOverview, PrintWriter writerDetail, String dataNamesString) {
    	writerOverview.println(";;;;Beendete Prozessinstanzen - Genaue Dauern;;;Beendete Prozessinstanzen - Dauern in Tagen;");
	    writerOverview.println("Prozess;Beendete Prozessinstanzen;Laufende Prozessinstanzen;Prozessinstanzen gesamt;" +
    	       "Krzeste Dauer;Durchschnittliche Dauer;Lngste Dauer;" +
    	       "Krzeste Dauer; Durchschnittliche Dauer; Lngste Dauer");
	    writerDetail.println("Prozess;Prozessinstanznr;Start;Ende;Dauer;Dauer in Tagen"+dataNamesString);
	}
	
	/*
	 * Removes all occurrences of Semicolons and Newlines and replaces them by Spaces.
	 * Semicolons and Newlines are interpreted as a delimiter in a csv file. Therefore the payload strings
	 * shouldn't contain these characters.
	 */
	private String removeUnwantedCharacters(String s){
		return s.replaceAll(";|\n", " ");
	}
	
	
    /*
     * Converts a duration (in seconds) to a string 
     * with days, hours, minutes, seconds in the format,
     * e.g. "23d 17:32:05"
     */
	private String formatduration(long durationInSeconds){
    	long days = durationInSeconds / 86400;
    	long restduration = durationInSeconds - (days * 86400);
    	long hours = restduration / 3600;	
    	restduration = restduration - (hours * 3600);
    	long minutes = restduration / 60;
    	long seconds = restduration - (minutes * 60);
    	
		return doubledigit(days) + "d " + doubledigit(hours) + ":" + doubledigit(minutes) + ":" + doubledigit(seconds);
	}
	
	
	/*
	 * Converts a duration (in seconds) to a string containing the
	 * duration in days with two decimal places
	 */
	private String durationInDays(long durationInSeconds){
		long days = durationInSeconds / 86400;
        double restduration = durationInSeconds / 86400.0 - days;
        long decimal = Math.round(restduration * 10000);
        String decimalString = Long.toString(decimal);
        while(decimalString.length()<4){
        	decimalString = "0" + decimalString;
        }
        return "" + days + "," + decimalString;
	}

	
	/*
	 * Returns aNumber as a string. If aNumber consists of only one digit,
	 * a leading "0" is added, so that it consists of two digits
	 * e.g.  2  -> "02"
	 */
	private String doubledigit(long aNumber){
		String numberString = ""+aNumber;
		if(aNumber<10){
			numberString = "0"+numberString;
		}
		return numberString;
	}

}
